
/*
 * Zeroth version and a half of a framework to get started.
 * A few ideas to take from this, but far too complicated
 * and not enough documentation.
 * First version to come when I got the time.
 *
 * Lionel Lemarie
 * hikey@playstation2-linux.com
 * 2002-07-01
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <math.h>
#include <ps2gs.h>
#include <ps2dma.h>
#include <linux/ps2/dev.h>
#include "packet.h"
#include "ps2hik.h"
#include "loaders.h"
#include "model.h"
#include "log.h"
#include "vu0.h"


void init();
void deinit(int);
int  main(int argc, char *argv[]);
void state_menu();
void state_load();
void state_game();
void state_quit();

void (*game_frame)();
int loop_again=1;

struct {
	image_t ps2tux;
	image_t boom;
	filelist_t filelist;
	scene_t scene;
	gifpacket_t gif_pkt;
} game_data;

#include <ps2vpu.h>
#include <ps2vpufile.h>
#include <ps2gs.h>

int main(int argc, char **argv)
{
	if (!ps2hik_init(argc,argv)) { printf("Error initialising devices, exiting...\n"); return 1; }
	ps2hik_set_deinit(deinit);

	log_init("log.txt", LOG_TRUNC(1), LOG_SYNC(0));

	ps2hik_printf_resolution();

	init();

	game_frame=state_menu;

	while (loop_again)
	{
		ps2hik_lock();
		game_frame();
		ps2hik_unlock();
	}

	ps2hik_release(0);

	return 0;
}

void init()
{
	init_image(&game_data.ps2tux);
	init_image(&game_data.boom);
	game_data.scene.nmodels=0;
	game_data.scene.models=NULL;
}

void deinit(int sig)
{
	printf("Deinitialising before quitting\n");
	free_image(&game_data.ps2tux);
	free_image(&game_data.boom);
	for (; game_data.scene.nmodels>0; game_data.scene.nmodels--)
		FreeRawModel(&game_data.scene.models[game_data.scene.nmodels-1]);
	free(game_data.scene.models);
}

void state_menu()
{
	game_frame=state_load;
}

void InitVIFPktStripped( model_t *mdl )
{
	int prim_type;
	int pre;
	int mesh_verts;

	int packet_size;
	int m, v, cnt1;

	int tags=0;

	packet_size=10;
	for (m=0; m<mdl->nmeshes; m++)
	{
		packet_size+=10;

		mesh_verts=mdl->mesh[m].nfaces+2;
		for (v=0; v<mesh_verts; v+=VU_VERTS_STRIP)
			packet_size+=VU_VERTS_STRIP+VU_VERTS_STRIP*2*3/4+4;
	}

	printf("-\nallocating %d qwords (%d bytes)\n", packet_size, packet_size*16);

	free(mdl->vpkt.buffer);
	mdl->vpkt.buffer=NULL;
	mdl->vpkt.buffer=memalign(64,16*packet_size);

	// init memory for data VIF PKT
	vifpacket_init( &mdl->vpkt.vp, mdl->vpkt.buffer );

	vifpacket_end( &mdl->vpkt.vp, 0 ); tags++;

	if (mdl->type==MODEL_BAD_FORMAT)
	{
		printf("bad format\n-\n");
		vifpacket_terminate( &mdl->vpkt.vp );
		return;
	}
	printf("good format\n");

	vifpacket_addcode( &mdl->vpkt.vp, 0x3000000 ); // set vu1 base
	vifpacket_addcode( &mdl->vpkt.vp, 0x2000200 ); // set vu1 offset

	vifpacket_opendirectcode(&mdl->vpkt.vp,0);
	vifpacket_addupkdata64(&mdl->vpkt.vp, GIF_SET_TAG(1,1,0,0,GIF_PACKED,4) );
	vifpacket_addupkdata64(&mdl->vpkt.vp, 0xEEEEL);
	vifpacket_addupkdata64(&mdl->vpkt.vp, PS2_GS_SETREG_ALPHA( 0,2,2,2,0xA0));
	vifpacket_addupkdata64(&mdl->vpkt.vp, PS2_GS_ALPHA_1);
	vifpacket_addupkdata64(&mdl->vpkt.vp, PS2_GS_SETREG_PABE( 1 ));
	vifpacket_addupkdata64(&mdl->vpkt.vp, PS2_GS_PABE);
	vifpacket_addupkdata64(&mdl->vpkt.vp, PS2_GS_SETREG_FBA( 0 ));
	vifpacket_addupkdata64(&mdl->vpkt.vp, PS2_GS_FBA_1);
	vifpacket_addupkdata64(&mdl->vpkt.vp, PS2_GS_SETREG_CLAMP(0, 0, 0, 0, 0, 0));
	vifpacket_addupkdata64(&mdl->vpkt.vp, PS2_GS_CLAMP_1);
	vifpacket_closedirectcode(&mdl->vpkt.vp);

	prim_type=PS2_GS_SETREG_PRIM( PS2_GS_PRIM_PRIM_TRISTRIP, GOURAUD(1), TEXTURED(0), 0, 0, 0, 0, 0, 0 );

	vifpacket_openupkcode( &mdl->vpkt.vp, ( 0 | 0x8000 ), V4_32, 4, 4 );

	mdl->local_screen=mdl->vpkt.vp.current;
	for (cnt1=0; cnt1<4; cnt1++) vifpacket_addupkdata128( &mdl->vpkt.vp, 0 );

	mdl->local_light=mdl->vpkt.vp.current;
	for (cnt1=0; cnt1<4; cnt1++) vifpacket_addupkdata128( &mdl->vpkt.vp, 0 );

	mdl->light_color=mdl->vpkt.vp.current;
	for (cnt1=0; cnt1<4; cnt1++) vifpacket_addupkdata128( &mdl->vpkt.vp, 0 );

	vifpacket_closeupkcode( &mdl->vpkt.vp );

	vifpacket_adddata( &mdl->vpkt.vp, VIF1_SET_MSCALF(0,0));

	for (m=0; m<mdl->nmeshes; m++)
	{
		// tex=GetTexture(mdl->mesh[m].texture);

		mesh_verts=mdl->mesh[m].nfaces+2;

		pre=1;

		for (v=0; v<mesh_verts; v+=VU_VERTS_STRIP)
		{
			int vu_verts=VU_VERTS_STRIP;

			if (v+VU_VERTS_STRIP>mesh_verts)
				vu_verts=mesh_verts%VU_VERTS_STRIP;

			vifpacket_openupkcode( &mdl->vpkt.vp, 0 | 0x8000, V4_32, 4, 4 );
			vifpacket_addupkdata64( &mdl->vpkt.vp, PS2_GIFTAG_SET_TOPHALF( NLOOP(vu_verts), EOP(1), PRE(pre), prim_type, PS2_GIFTAG_FLG_PACKED, 2 ) );
			vifpacket_addupkdata64( &mdl->vpkt.vp, 0x41L );
			vifpacket_closeupkcode( &mdl->vpkt.vp );

			vifpacket_openupkcode( &mdl->vpkt.vp, 1 | 0x8000, V3_32, 4, 4 );
			vifpacket_addupkdata128N(&mdl->vpkt.vp, (u_long128*)(mdl->mesh[m].vertex+v*VERT_DATA_LEN_STRIP), VU_VERTS_STRIP*3/4);
			vifpacket_closeupkcode( &mdl->vpkt.vp );

			vifpacket_openupkcode( &mdl->vpkt.vp, 101 | 0x8000, V3_32, 4, 4 );
			vifpacket_addupkdata128N(&mdl->vpkt.vp, (u_long128*)(mdl->mesh[m].normal+v*VERT_DATA_LEN_STRIP), VU_VERTS_STRIP*3/4);
			vifpacket_closeupkcode( &mdl->vpkt.vp );

			vifpacket_openupkcode( &mdl->vpkt.vp, 201 | 0x8000, V4_32, 4, 4 );
			for (cnt1=0; cnt1<VU_VERTS_STRIP; cnt1++)
				vifpacket_addupkdata128(&mdl->vpkt.vp, *(u_long128*)&mdl->mesh[m].rgb);
			vifpacket_closeupkcode( &mdl->vpkt.vp );

			vifpacket_addcode( &mdl->vpkt.vp, VIF1_SET_MSCNT( 0 ) );

			pre=0;
		}
	}

	//vifpacket_end( &mdl->vpkt.vp, 0 ); tags++;

	vifpacket_terminate( &mdl->vpkt.vp );

	printf("finished packet: %d qwords, %d tags\n-\n", vifpacket_size(&mdl->vpkt.vp), tags);
}

void state_load()
{
	// Load PCX files
	load_image_pcx("ps2tux.pcx", &game_data.ps2tux);
	load_image_pcx("boom.pcx", &game_data.boom);

	// Get an address for them in VRAM
	ps2hik_set_image_address(&game_data.ps2tux);
	ps2hik_set_image_address(&game_data.boom);

	// load the list of the models to load
	GetFileList("list.txt", &game_data.filelist);

	// get the number of models
	game_data.scene.nmodels=game_data.filelist.name_count;

	// if there are any, build a list for each
	if (game_data.scene.nmodels)
	{
		// model counter
		int m;
		// used for placement calculations
		float ftotal=game_data.scene.nmodels-1;
		float fww;

		// allocate memory for the model structures
		// important to align because the first data in there is the transform variables
		game_data.scene.models=(model_t *)memalign(16, game_data.scene.nmodels*sizeof(model_t));
		game_data.scene.nfaces=0;

		// for each model
		for (m=0; m<game_data.scene.nmodels; m++)
		{
			// initialise the model
			InitRawModel(game_data.scene.models+m);
			// load the data
			LoadFileToRawModel(game_data.filelist.names[m], game_data.scene.models+m);
			// build the vif packet
			InitVIFPktStripped(game_data.scene.models+m);

			// 
			if (game_data.scene.models[m].type!=MODEL_BAD_FORMAT)
			{
				game_data.scene.nfaces+=game_data.scene.models[m].nfaces;
				memcpy(game_data.scene.models[m].obj_rot0, game_data.filelist.rots[m], sizeof(ps2_vu0_fvector));
				fww=m-ftotal/2;
				fww/=ftotal+1;
				game_data.scene.models[m].obj_trans[0]=1200.0f*fww;

			//	vifpacket_dump(&game_data.scene.models[m].vpkt.vp);
			//	exit(1);
			}
		}
	}

	printf("%d models loaded, %d polys\n", game_data.scene.nmodels, game_data.scene.nfaces);

	// setting up a sprite for the background
	{
		struct 
		{
			int r, g, b, a;
			int x0, y0;
			int x1, y1;
			int w;
			int h;
			int x;
			int y;
		} sprite;

		u_int *base_ptr=NULL;

		// allocate some memory, 16k is more than enough
		base_ptr=(u_int *)memalign(16, 16*1024);

		// full screen
		sprite.x = 0 << 4;
		sprite.y = 0 << 4;
		sprite.w = gparam->width << 4;
		sprite.h = gparam->height << 4;

		// full brightness
		sprite.r = 0x80;
		sprite.g = 0x80;
		sprite.b = 0x80;
		sprite.a = 0x80;

		// tranform to screen coordinate system
		sprite.x0 = (((4096-gparam->width)/2)<<4) + sprite.x;
		sprite.y0 = (((4096-gparam->height)/2)<<4) + sprite.y;
		sprite.x1 = sprite.x0 + sprite.w;
		sprite.y1 = sprite.y0 + sprite.h;

		// start GIF packet for the sprite
		gifpacket_init(&game_data.gif_pkt, (u_long128*)base_ptr);

		// Add GIF tag to setup enviromnent
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GIFTAG_SET_TOPHALF( NLOOP(1), EOP(1), PRE(0), 0, PS2_GIFTAG_FLG_PACKED, NREG(3) ) );
		gifpacket_addgsdata( &game_data.gif_pkt, 0xEEEL );

		// Set TEX0 register to use the ps2tux image
		gifpacket_addgsdata( &game_data.gif_pkt,
				PS2_GS_SETREG_TEX0( game_data.ps2tux.texture.fbp, game_data.ps2tux.texture.fbw,
					game_data.ps2tux.texture.psm, game_data.ps2tux.tw, game_data.ps2tux.th, 1,
					PS2_GS_TEX_TFX_DECAL, game_data.ps2tux.clut.fbp, game_data.ps2tux.clut.psm, 0, 0, 1 ) );
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_TEX0_1 );

		// Set TEX1 to bilinear filtering
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_SETREG_TEX1_1( 0, 0, PS2_GS_TEX1_MMAG_LINEAR, PS2_GS_TEX1_MMIN_LINEAR, 0, 0, 0 ) );
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_TEX1_1 );

		// Set CLAMP to clamp, not repeat
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_SETREG_CLAMP( 1, 1, 0, game_data.ps2tux.texture.w, 0, game_data.ps2tux.texture.h) );
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_CLAMP_1 );


		// Add GIF tag to send vertex data
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GIFTAG_SET_TOPHALF( NLOOP(1), EOP(1), PRE(0), 0, PS2_GIFTAG_FLG_REGLIST, NREG(6) ) );
		gifpacket_addgsdata( &game_data.gif_pkt, 0x525210L );

		// add the sprite data
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_SETREG_PRIM( PS2_GS_PRIM_PRIM_SPRITE, GOURAUD(0), TEXTURED(1), 0, 0, 0, 0, 0, 0 ) );
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_SETREG_RGBAQ( sprite.r, sprite.g, sprite.b, sprite.a, FLOAT_ONE ) );
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_SETREG_ST( 0, 0 ) );
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_SETREG_XYZ( sprite.x0, sprite.y0, 1 ) ); 
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_SETREG_ST( FLOAT_ONE, FLOAT_ONE ) );
		gifpacket_addgsdata( &game_data.gif_pkt, PS2_GS_SETREG_XYZ( sprite.x1, sprite.y1, 1 ) );

		// terminate the packet
		gifpacket_terminate( &game_data.gif_pkt );

		// gifpacket_dump(&game_data.gif_pkt);
		// printf("\n");
	}

	// loading is finished, switch to game loop
	game_frame=state_game;
}








// Game loop
void state_game()
{

	// handles for VU1 and GS
	extern int fd_gs;
	extern ps2_vpu *vpu1_handle;

	// stuff we need to calculate the tranform matrices
	static ps2_vu0_fvector camera_p = { 0.0f, 0.0f, -1000.0f, 0.0f };
	static ps2_vu0_fvector camera_zd = { 0.0f, 0.0f, 1.0f, 1.0f };
	static ps2_vu0_fvector camera_yd = { 0.0f, 1.0f, 0.0f, 1.0f };
	static ps2_vu0_fvector camera_rot = { 0, 0, 0, 0 };
	static ps2_vu0_fvector light0 = { 1.0f, 0.0f, 0.3f, 0.0f };
	static ps2_vu0_fvector light1 = { 0.0f, 1.0f, 0.3f, 0.0f };
	static ps2_vu0_fvector light2 = { 0.0f, 0.0f, 1.0f, 0.0f };
	static ps2_vu0_fvector color0 = { 0.4f, 0.4f, 0.4f, 1.0f };
	static ps2_vu0_fvector color1 = { 0.4f, 0.4f, 0.4f, 1.0f };
	static ps2_vu0_fvector color2 = { 0.4f, 0.4f, 0.4f, 1.0f };
	static ps2_vu0_fvector ambient = { 0.2f, 0.2f, 0.2f, 0.0f };
	static ps2_vu0_fmatrix local_world;
	static ps2_vu0_fmatrix world_view;
	static ps2_vu0_fmatrix view_screen;
	static ps2_vu0_fmatrix local_screen;
	static ps2_vu0_fmatrix normal_light;
	static ps2_vu0_fmatrix light_color;
	static ps2_vu0_fmatrix local_light;
	static ps2_vu0_fmatrix work;

	// model counter
	int mm;


	// Send the GIF packet for the background sprite
	ps2_dma_start_n(fd_gs, game_data.gif_pkt.base, gifpacket_size(&game_data.gif_pkt));

	// view -> screen matrix
	ps2_vu0_view_screen_matrix(view_screen, 512.0f, 1.0f, gparam->pixel_ratio,
			gparam->center_x, gparam->center_y, 1.0f, ((1 << 24) - 1), 1.0f, 65536.0f);

	// color&normal matrix setting
	ps2_vu0_normal_light_matrix(normal_light, light0, light1, light2);

	// world -> view matrix
	vu0_rot_camera_matrix(world_view, camera_p, camera_zd, camera_yd, camera_rot);

	// light color
	ps2_vu0_light_color_matrix(light_color, color0, color1, color2, ambient);

	// for each model
	for (mm=0; mm<game_data.scene.nmodels; mm++)
	{
		// Rotate the model
		game_data.scene.models[mm].obj_rot[1]+=M_PI/150.0f;
		// Rotation must be between -PI and PI, otherwise it does weird things
		if (game_data.scene.models[mm].obj_rot[1]>M_PI) game_data.scene.models[mm].obj_rot[1]-=2.0f*M_PI;

		// local -> world (rotate)matrix
		ps2_vu0_unit_matrix(work);
		ps2_vu0_rot_matrix(work, work, game_data.scene.models[mm].obj_rot0);
		ps2_vu0_rot_matrix(local_world, work, game_data.scene.models[mm].obj_rot);

		// light(normal) -> local_light matrix
		ps2_vu0_mul_matrix(local_light, normal_light, local_world);

		// local -> world (rotate&translate)matrix
		ps2_vu0_trans_matrix(local_world, local_world, game_data.scene.models[mm].obj_trans);

		// local -> screen matrix
		ps2_vu0_mul_matrix(work, world_view, local_world);
		ps2_vu0_mul_matrix(local_screen, view_screen, work);

		// if we got the pointers for the matrices in the dma list
		if (game_data.scene.models[mm].local_screen!=NULL)
		{
			// copy the tranform matrices into the dma list
			memcpy(game_data.scene.models[mm].local_screen, local_screen, sizeof(ps2_vu0_fmatrix));
			memcpy(game_data.scene.models[mm].local_light, local_light, sizeof(ps2_vu0_fmatrix));
			memcpy(game_data.scene.models[mm].light_color, light_color, sizeof(ps2_vu0_fmatrix));
		}

		// send it baby
		ps2_dma_start(ps2_vpu_fd(vpu1_handle), NULL, (ps2_dmatag*)game_data.scene.models[mm].vpkt.vp.base);
	}

	// This is the end of the frame, we send the textures for the next frame.
	ps2_gs_load_image((struct ps2_image*)&game_data.ps2tux.clut);
	ps2_gs_load_image((struct ps2_image*)&game_data.ps2tux.texture);
	ps2_gs_load_image((struct ps2_image*)&game_data.boom.clut);
	ps2_gs_load_image((struct ps2_image*)&game_data.boom.texture);
}

void state_quit()
{
	// bye bye
	printf("Quit requested by user\n");
	loop_again=0;
}








