#ifndef __STRIPPED_H__
#define __STRIPPED_H__

#include "ps2gs.h"
#include "packet.h"
#include "vu0.h"

//#define FILEHOST "host0:"
#define FILEHOST ""

#define TRIMODEL_VERSION (1)
#define STRIPPED_VERSION (6)
#define STRIPPED_MAGIC ("STRIPPED")

#define MODEL_BAD_FORMAT (0)
#define MODEL_TRIMODEL   (1)
#define MODEL_STRIPPED   (2)

#define VERT_DATA_LEN_STRIP (3*4)
#define NORM_DATA_LEN_STRIP (3*4)
#define ST_DATA_LEN_STRIP   (2*4)
#define VERT_DATA_LEN_TRI (16)

#define VU_VERTS_STRIP (100)

#define _SET_VECTOR(_p,_x,_y,_z,_w)  (_p[0] = _x, _p[1] = _y, _p[2] = _z, _p[3] = _w)
#define MULQWORD(a, b, c) ( \
		((qword_t*)&a)->f32[0]=((qword_t*)&b)->f32[0]*((qword_t*)&c)->f32[0], \
		((qword_t*)&a)->f32[1]=((qword_t*)&b)->f32[1]*((qword_t*)&c)->f32[1], \
		((qword_t*)&a)->f32[2]=((qword_t*)&b)->f32[2]*((qword_t*)&c)->f32[2], \
		((qword_t*)&a)->f32[3]=((qword_t*)&b)->f32[3]*((qword_t*)&c)->f32[3])
#define ADDQWORD(a, b, c) ( \
		((qword_t*)&a)->f32[0]=((qword_t*)&b)->f32[0]+((qword_t*)&c)->f32[0], \
		((qword_t*)&a)->f32[1]=((qword_t*)&b)->f32[1]+((qword_t*)&c)->f32[1], \
		((qword_t*)&a)->f32[2]=((qword_t*)&b)->f32[2]+((qword_t*)&c)->f32[2], \
		((qword_t*)&a)->f32[3]=((qword_t*)&b)->f32[3]+((qword_t*)&c)->f32[3])
#define CPYQWORD(a, b) ( \
		((qword_t*)&a)->f32[0]=((qword_t*)&b)->f32[0], \
		((qword_t*)&a)->f32[1]=((qword_t*)&b)->f32[1], \
		((qword_t*)&a)->f32[2]=((qword_t*)&b)->f32[2], \
		((qword_t*)&a)->f32[3]=((qword_t*)&b)->f32[3])
#define NORM_VECTOR(a) { \
		float l= \
			((qword_t*)&a)->f32[0]*((qword_t*)&a)->f32[0]+ \
			((qword_t*)&a)->f32[1]*((qword_t*)&a)->f32[1]+ \
			((qword_t*)&a)->f32[2]*((qword_t*)&a)->f32[2]; \
		if (l==0.0f) l=1.0f; \
		l=1.0f/l; \
		((qword_t*)&a)->f32[0]*=l; \
		((qword_t*)&a)->f32[0]*=l; \
		((qword_t*)&a)->f32[0]*=l; \
	}

typedef union
{
	u_long128     ul128;
	u_long        ul64[2];
	u_int         ui32[4];
	float         f32[4];
	ps2_vu0_fvector fvect;
} qword_t;

typedef struct
{
	vifpacket_t vp;
	u_long128 *buffer;
	unsigned int buffer_size;
} vifpkt_t;

typedef struct
{
	qword_t rgb;
	unsigned int tex_index;
	unsigned int nfaces;
	int texture;
	char *vertex;
	char *normal;
	char *st;
} mesh_t;

typedef struct
{
	ps2_vu0_fvector obj_trans;
	ps2_vu0_fvector obj_rot;
	ps2_vu0_fvector obj_rot0;
	u_int *local_screen;
	u_int *local_light;
	u_int *light_color;
	u_int *local_clip;
	u_int *outline_factor;
	unsigned int nfaces;
	unsigned int nmeshes;
	int type;
	mesh_t *mesh;
	vifpkt_t vpkt;
} model_t;

typedef struct
{
	unsigned int nfaces;
	unsigned int nmodels;
	model_t *models;
} scene_t;

typedef struct
{
	int name_count, index;
	int allocated;
	char **names;
	ps2_vu0_fvector *rots;
} filelist_t;

extern int loading;

void GetFileList(const char*filename, filelist_t *list);
void InitRawModel(model_t *mdl);
void LoadFileToRawModel(char *filename, model_t *mdl);
void LoadStrippedToRawModel(char *filename, model_t *mdl);
void FreeRawModel(model_t *mdl);

#endif

